#!/usr/bin/python3
"""
Copyright (c) 2025 Penterep Security s.r.o.

ptinsearcher - Web / File information extractor

ptinsearcher is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

ptinsearcher is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with ptinsearcher. If not, see <https://www.gnu.org/licenses/>.
"""

import argparse
import os
import re
import sys; sys.path.append(__file__.rsplit("/", 1)[0])
import urllib
import requests
from concurrent.futures import ThreadPoolExecutor
import threading

from ptlibs import ptmisclib, ptjsonlib, ptprinthelper
from modules import _website_scraper, _files_scraper, metadata, results

from modules._website_scraper import WebsiteScraper
from modules._files_scraper import FileScraper
from modules.results.results import print_result

from modules.cli.extract_flags import parse_extract_flags
from modules.cli.args_validator import validate_args

from _version import __version__

class PtInsearcher:
    def __init__(self, args):
        self.ptjsonlib                = ptjsonlib.PtJsonLib()
        self.url_list                 = self._get_url_list(args)
        self.args                     = args
        self.things_to_extract: dict  = args.extract
        self.use_json                 = args.json
        self.group_parameters         = args.group_parameters
        self.without_parameters       = args.without_parameters
        self.file_handler             = open(args.output, "w") if args.output and not args.output_parts else None
        self.lock                     = threading.Lock()

    def run(self, args: argparse.Namespace):
        if len(self.url_list) == 1:
            self.args.threads = 1
        else:
            self.args.grouping_complete = True


        # Paralell processing of targets
        result_list: list = [result for result in self.process_targets(targets=self.url_list)]


        # Print results if grouping is enabled or multiple results with grouping_complete - for summary
        if (self.args.grouping) or (len(result_list) > 1 and self.args.grouping_complete):
            print_result(result_list, args)

        self.ptjsonlib.set_status("finished")
        ptprinthelper.ptprint(self.ptjsonlib.get_result_json(), "", self.use_json)

    def process_targets(self, targets: list[str]):
        """
        Processes multiple targets concurrently using ThreadPoolExecutor.

        Parameters:
            targets (list[str]): A list of URLs or file paths to process.

        Returns:
            list: A list of results from processing each target.
        """
        results = []

        with ThreadPoolExecutor(max_workers=self.args.threads) as executor:
            future_to_target = {executor.submit(self.process_target, target): target for target in targets}
            for future in future_to_target:
                try:
                    result = future.result()
                    if result:
                        with self.lock:
                            results.append(result)
                            if not self.args.grouping: # if not -g
                                print_result([result], self.args) # Print output continuously
                except Exception as e:
                    ptprinthelper.ptprint(f"Error processing {future_to_target[future]}: {e}", "ERROR")
        return results

    def process_target(self, target: str):
        """
        Processes the given target, identifying it as either a URL or a file path.

        This method checks if the `target` is a valid URL or an existing file.
        It then processes the target appropriately:
        - For URLs: Scrapes the website using `self.website_scraper`.
        - For files: Processes the file using `scrape_files.FileScraper`.

        Parameters:
            target (str): The source to be processed, which can be a URL or a local file path.

        Returns:
            Any: The result of scraping or file processing.

        Raises:
            ValueError: If the target does not represent a valid URL or file and error handling is triggered.
        """
        if self._target_is_url(target):
            scrape_result = WebsiteScraper(args=self.args, things_to_extract=self.things_to_extract, url_list_len=len(self.url_list), ptjsonlib=self.ptjsonlib, lock=self.lock).scrape_url(url=target)
            return scrape_result

        elif self._target_is_file(target):
            scrape_result = FileScraper(args=self.args, things_to_extract=self.things_to_extract, ptjsonlib=self.ptjsonlib, lock=self.lock).process_file(path_to_local_file=os.path.abspath(target))
            return scrape_result

        else:
            error_message = f"{target} is neither an existing file nor a valid URL."
            if len(self.url_list) > 1:
                ptprinthelper.ptprint(error_message, "ERROR")
                return
            self.ptjsonlib.end_error(error_message, self.use_json)

    def _target_is_file(self, source: str) -> bool:
        """Check whether the provided source is an existing file"""
        return os.path.isfile(source)

    def _target_is_url(self, source: str) -> bool:
        """Check whether the provided source is a valid URL"""
        regex = re.compile(
        r'^(?:http|ftp)s?://' # http:// or https://
        r'(?:(?:[A-Z0-9](?:[A-Z0-9-]{0,61}[A-Z0-9])?\.)+(?:[A-Z]{2,6}\.?|[A-Z0-9-]{2,}\.?)|' #domain...
        r'localhost|' #localhost...
        r'\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3})'
        r'(?::\d+)?' # optional ports
        r'(?:/?|[/?]\S+)$', re.IGNORECASE)
        return regex.match(source)

    def _get_url_list(self, args) -> list:
        """Generates a list of URLs based on the provided arguments."""
        if args.url:
            return args.url
        elif args.file:
            try:
                url_list = self._read_file(args.file, args.domain)
                if args.extension_yes:
                   url_list = list(dict.fromkeys([url for url in url_list if url.endswith(tuple(args.extension_yes))]))
                if args.extension_no:
                    url_list = list(dict.fromkeys([url for url in url_list if not url.endswith(tuple(args.extension_no))]))
                return list(dict.fromkeys(url_list))
            except Exception as e:
                self.ptjsonlib.end_error(f"Error reading from provided file. ({e})", args.json)

    def _read_file(self, filepath, domain) -> list:
        domain = self._normalize_domain(domain) if domain else None
        target_list = []
        try:
            with open(filepath, "r") as f:
                for line in f:
                    line = line.strip("\n")
                    if domain:
                        path = urllib.parse.urlparse(line).path
                        while path.startswith("/"): path = path[1:]
                        while path.endswith("/"): path = path[:-1]
                        if not path: continue
                        target_list.append(domain + path)
                    else:
                        if re.match(r'https?://', line):
                            while line.endswith("/"): line = line[:-1]
                            target_list.append(line)
            return target_list
        except FileNotFoundError:
            self.ptjsonlib.end_error(f"File not found {os.path.join(os.getcwd(), self.args.file)}", self.use_json)

    def _normalize_domain(self, domain: str) -> str:
        """Adjusts provided <domain>"""
        o = urllib.parse.urlparse(domain)
        if not re.match("http[s]?$", o.scheme):
            self.ptjsonlib.end_error(f"Missing or invalid scheme, supported schemes are: [HTTP, HTTPS]", self.use_json)
        return domain + "/" if not o.path.endswith("/") else domain


def get_help():
    return [
        {"description": ["Source information extractor"]},
        {"usage": ["ptinsearcher <options>"]},
        {"usage_example": [
            "ptinsearcher -u https://www.example.com/",
            "ptinsearcher -u https://www.example.com/ --extract E        # Extract emails",
            "ptinsearcher -u https://www.example.com/ --extract UQX      # Extract internal URLs, internal URLs w/ parameters, external URLs",
            "ptinsearcher -f url_list.txt --grouping                     ",
            "ptinsearcher -f url_list.txt --grouping-complete            ",
            "ptinsearcher -f urls.txt -e R -s \"text[0-9]*\" -gc",
            "ptinsearcher -f url_list.txt",
            "ptinsearcher -u image.jpg -e M",
            "ptinsearcher -u images/*.jpg -e M",
        ]},
        {"options": [
            ["-u",  "--url",                    "<url>",            "Test URL or File"],
            ["-f",  "--file",                   "<file>",           "Load list of URLs from file"],
            ["-d",  "--domain",                 "<domain>",         "Domain - merge domain with filepath. Use when wordlist contains filepaths (e.g. /index.php)"],
            ["-e",  "--extract",                "<extract>",        "Specify data to extract:"],
            ["",    "",                         "   A",             "  All (extracts everything - default option)"],
            ["",    "",                         "   E",             "  Emails"],
            ["",    "",                         "   S",             "  Subdomains"],
            ["",    "",                         "   C",             "  Comments"],
            ["",    "",                         "   F",             "  Forms"],
            ["",    "",                         "   I",             "  IP addresses"],
            ["",    "",                         "   P",             "  Phone numbers"],
            ["",    "",                         "   U",             "  Internal urls"],
            ["",    "",                         "   Q",             "  Internal urls with parameters"],
            ["",    "",                         "   X",             "  External urls"],
            ["",    "",                         "   N",             "  Insecure urls"],
            ["",    "",                         "   M",             "  Metadata"],
            ["",    "",                         "   R",             "  Regex"],
            ["",    "",                         "   K",             "  Google Keys"],
            ["",    "",                         "",                 ""],
            ["-s",  "--search",                 "<re>",             "Search for provided regex."],
            ["-ey", "--extension-yes",          "<extensions>",     "Process only URLs from <list> that end with <extension-yes>"],
            ["-en", "--extension-no",           "<extensions>",     "Process only URLs from <list> that do not end with <extension-no>"],
            ["-g",  "--grouping",               "",                 "Group findings from multiple sources into one table"],
            ["-gp", "--group-parameters",       "",                 "Group URL parameters"],
            ["-wp", "--without-parameters",     "",                 "Without URL parameters"],
            ["-op", "--output-parts",           "",                 "Save each extract-type to separate file"],
            ["-o",  "--output",                 "<output>",         "Save output to file"],
            ["-vv",  "--very-verbose",          "",                 "Enable very verbose output"],

            ["-p",  "--proxy",                  "<proxy>",          "Set proxy (e.g. http://127.0.0.1:8080)"],
            ["-T",  "--timeout",                "<timeout>",        "Set timeout"],
            ["-c",  "--cookie",                 "<cookie=value>",   "Set cookie"],
            ["-a",  "--user-agent",             "<user-agent>",     "Set User-Agent"],
            ["-t",  "--threads",                "<threads>",        "Set Threads"],
            ["-H",  "--headers",                "<header:value>",   "Set custom header(s)"],
            ["-r",  "--redirects",              "",                 "Follow redirects (default False)"],
            ["-C",  "--cache",                  "",                 "Cache requests (load from tmp in future)"],
            ["-v",  "--version",                "",                 "Show script version and exit"],
            ["-h",  "--help",                   "",                 "Show this help message and exit"],
            ["-j",  "--json",                   "",                 "Output in JSON format"],
        ]
        }]


def parse_args():
    parser = argparse.ArgumentParser(add_help=False, usage=f"{SCRIPTNAME} <options>")
    parser.add_argument("-u",  "--url",                type=str, nargs="+")
    parser.add_argument("-f",  "--file",               type=str)
    parser.add_argument("-d",  "--domain",             type=str)
    parser.add_argument("-e",  "--extract",            type=str, default="A")
    parser.add_argument("-ey", "--extension-yes",      type=str, nargs="+")
    parser.add_argument("-en", "--extension-no",       type=str, nargs="+")
    parser.add_argument("-pd", "--post-data",          type=str)
    parser.add_argument("-o",  "--output",             type=str)
    parser.add_argument("-s",  "--search",             type=str)
    parser.add_argument("-p",  "--proxy",              type=str)
    parser.add_argument("-T",  "--timeout",            type=int)
    parser.add_argument("-t",  "--threads",            type=int, default=5)
    parser.add_argument("-c",  "--cookie",             type=str, nargs="+")
    parser.add_argument("-a",  "--user-agent",         type=str, default="Penterep Tools")
    parser.add_argument("-H",  "--headers",            type=ptmisclib.pairs, nargs="+")

    group = parser.add_mutually_exclusive_group(required=False)
    group.add_argument("-g",  "--grouping",           action="store_true")
    group.add_argument("-gc", "--grouping-complete",  action="store_true")

    parser.add_argument("-r",  "--redirects",          action="store_true")
    parser.add_argument("-op", "--output-parts",       action="store_true")
    parser.add_argument("-gp", "--group-parameters",   action="store_true")
    parser.add_argument("-wp", "--without-parameters", action="store_true")
    parser.add_argument("-vv",  "--very-verbose",          action="store_true")

    parser.add_argument("-C",  "--cache-requests",     action="store_true")
    parser.add_argument("-j",  "--json",               action="store_true")
    parser.add_argument("-v",  "--version",            action="version", version=f"%(prog)s {__version__}", help="show version")

    parser.add_argument("--socket-address",          type=str, default=None)
    parser.add_argument("--socket-port",             type=str, default=None)
    parser.add_argument("--process-ident",           type=str, default=None)


    if len(sys.argv) == 1 or "-h" in sys.argv or "--help" in sys.argv:
        ptprinthelper.help_print(get_help(), SCRIPTNAME, __version__)
        sys.exit(0)

    args = parser.parse_args()
    _ptjsonlib = ptjsonlib.PtJsonLib()

    args.extract: dict = parse_extract_flags(
        args=args,
        extract_str=args.extract,
        on_error=lambda msg: _ptjsonlib.end_error(msg, args.json)
        )

    ptprinthelper.print_banner(SCRIPTNAME, __version__, args.json, space=0)
    return args


def main():
    global SCRIPTNAME
    SCRIPTNAME = "ptinsearcher"
    requests.packages.urllib3.disable_warnings()
    args = parse_args()
    script = PtInsearcher(args)
    script.run(args)


if __name__ == "__main__":
    main()



"""
CLI argument validation logic.

Ensures mutually exclusive or required arguments are properly used
and triggers formatted JSON errors if rules are violated.
"""


def validate_args(args, ptjsonlib):
    """
    Validates combinations and presence of CLI arguments.

    Raises a JSON-formatted error via ptjsonlib if invalid usage is detected.
    """

    if len(args.url_list) > 1 and args.json:
        ptjsonlib.end_error("Cannot test more than 1 domain while --json parameter is present", "ERROR")
    if (args.grouping or args.grouping_complete) and args.json:
        ptjsonlib.end_error("Cannot use -g or -gc parameters while --json parameter is present", args.json)
    if args.grouping and args.grouping_complete:
        ptjsonlib.end_error("Cannot use both -g and -gc parameters together", args.json)

    if args.output_parts and not args.output:
        ptjsonlib.end_error("Missing --output parameter", args.json)
    if args.input and args.file:
        ptjsonlib.end_error("Cannot use --input and --list parameter together", args.json)
    if args.domain and not args.file:
        ptjsonlib.end_error("--list required to use with --domain", args.json)
    if not args.input and not args.file:
        ptjsonlib.end_error("--input or --list parameter required", args.json)
    if (args.extension_yes or args.extension_no) and not args.file:
        ptjsonlib.end_error("--list required for usage of --extension-yes / --extension-no parameters", args.json)
    if args.extension_yes and args.extension_no:
        ptjsonlib.end_error("Cannot combine --extension-yes together with --extension-no", args.json)



# search_parser.py

import re
from requests import Response

def parse_search_regex(response: Response, search_regex: str) -> list[str]:
    """
    Searches the HTTP response using exactly the regex provided in `search_regex`.

    Args:
        response (Response): The HTTP response object from `requests`.
        search_regex (str): Regex pattern from args.search.

    Returns:
        list[str]: A list of unique matches found.
    """
    try:
        pattern = re.compile(search_regex)
    except re.error as e:
        raise ValueError(f"Invalid regex: {e}") from e

    matches = pattern.findall(response.text)

    # if the regex contains groups → tuples → flatten
    cleaned_matches = []
    for m in matches:
        if isinstance(m, tuple):
            cleaned_matches.extend([x for x in m if x])
        else:
            cleaned_matches.append(m)

    return sorted(set(cleaned_matches))



import re
import requests
import urllib
import html
import sys

from bs4 import BeautifulSoup

from ptlibs import ptprinthelper, ptmisclib, ptnethelper

from modules import metadata, emails, comments, forms, phone_numbers, ip_addresses, urls
from modules.scraping import google, re_search


from modules.results.base_result import BaseResult

class WebsiteScraper:
    def __init__(self, args, things_to_extract: dict, url_list_len: int, ptjsonlib: object, lock: object):
        self.use_json: bool = args.json
        self.extract_types: dict = things_to_extract
        self.ptjsonlib: object = ptjsonlib
        self.url_list_len = url_list_len
        self.args = args
        self.lock = lock

    def scrape_url(self, url: str) -> dict:
        """Scrape <things_to_extract> from <url>'s response"""
        self.result = BaseResult(source=url)
        self.result.data: dict = {"url": '', "metadata": {}, "emails": [], "phone_numbers": [], "ip_addresses": [], "abs_urls": [], "internal_urls": [], "internal_urls_with_parameters": [], "external_urls": [], "insecure_sources": [], "subdomains": [], "forms": [], "comments": {}}
        try:
            response = self.get_response(url)
            self.result.data["url"] = response.url
            if response is None:
                return self.result

            if self.stop_on_redirect(response=response):
                return self.result

            if response.text:
                return self._scrape_response(response)

            #if response.url.endswith("robots.txt"):
            #    return self.parse_robots_txt(response)
            else:
                self.result.logs.append(["Response returned no content", "ERROR"])
                return self.result

        except requests.exceptions.RequestException:
            self.result.logs.append(["Server not responding", "ERROR"])
            return self.result

    def target_is_url(self, source: str) -> bool:
        """Check whether the provided source is a valid URL"""
        regex = re.compile(
        r'^(?:http|ftp)s?://' # http:// or https://
        r'(?:(?:[A-Z0-9](?:[A-Z0-9-]{0,61}[A-Z0-9])?\.)+(?:[A-Z]{2,6}\.?|[A-Z0-9-]{2,}\.?)|' #domain...
        r'localhost|' #localhost...
        r'\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3})'
        r'(?::\d+)?' # optional ports
        r'(?:/?|[/?]\S+)$', re.IGNORECASE)
        return regex.match(source)

    def get_response(self, url: str):
        """Retrieves response from <url>"""
        self.result.logs.append([f"Request URL: {url}", "TITLE"])
        response = ptmisclib.load_url_from_web_or_temp(url=url, method="GET" if not self.args.post_data else "POST", headers=ptnethelper.get_request_headers(self.args), proxies={"http": self.args.proxy, "https": self.args.proxy}, data=self.args.post_data, timeout=self.args.timeout, redirects=self.args.redirects, verify=False, cache=self.args.cache_requests)
        response.encoding = response.apparent_encoding

        self.result.logs.append([f"Content-Type: {response.headers.get('content-type', 'None').split(';')[0]}", "TITLE"])
        return response

        """
        except requests.exceptions.RequestException:
            ptprinthelper.ptprint(f"[error]", condition=not self.use_json)
            if self.url_list_len > 1:
                ptprinthelper.ptprint(f"Server not responding", "ERROR")
                return
            else:
                self.ptjsonlib.end_error("Server not responding", self.use_json)
        """

    def stop_on_redirect(self, response):
        """Stop on redirect if not --redirects"""
        history = [*response.history, response] if response.history else [response]
        if response.is_redirect and not self.args.redirects:
            if response.headers.get("location"):
                self.result.logs.append([f"Location: {response.headers.get('location')}", "TITLE"])

            self.result.logs.append([f"Redirects disabled, use -r/--redirect to follow", "ERROR"])
            return self.result

        if self.args.redirects and len(history) > 1:
            self.result.logs.append([f"Redirect URL: {history[-1].url} [{history[-1].status_code}]", "TITLE"])

    def _get_soup(self, response):
        if "<!ENTITY".lower() in response.text.lower():
            ptprinthelper.ptprint(f"Forbidden entities found", "ERROR", not self.use_json, colortext=True)
            return False
        else:
            soup = BeautifulSoup(response.text, features="lxml")
            bdos = soup.find_all("bdo", {"dir": "rtl"})
            for item in bdos:
                item.string.replace_with(item.text[::-1])
            return soup

    def _scrape_response(self, response) -> dict:
        """Extracts <extract_types> from <response>"""
        content_type = response.headers.get("content-type")

        # Find Metadata
        if self.extract_types["metadata"]:
            self.result.data["metadata"] = metadata.MetadataExtractor(args=self.args, ptjsonlib=self.ptjsonlib).get_metadata(response=response)
            self.ptjsonlib.add_node(self.ptjsonlib.create_node_object("metadata", None, None, properties={"metadata": self.result.data["metadata"]}))

        if "text" not in content_type:
            return self.result

        PAGE_CONTENT = urllib.parse.unquote(urllib.parse.unquote(html.unescape(response.text)))

        # Find Emails
        if self.extract_types["emails"]:
            self.result.data["emails"] = emails.find_emails(PAGE_CONTENT)
            self.ptjsonlib.add_node(self.ptjsonlib.create_node_object("emails", None, None, properties={"emails": self.result.data["emails"]}))

        # Find Page Comments
        if self.extract_types["comments"]:
            self.result.data["comments"] = comments.find_comments(PAGE_CONTENT)
            self.ptjsonlib.add_node(self.ptjsonlib.create_node_object("comments", None, None, properties={"comments": self.result.data["comments"]}))

        # Find Phone numbers
        if self.extract_types["phone_numbers"]:
            self.result.data["phone_numbers"] = phone_numbers.find_phone_numbers(PAGE_CONTENT)
            self.ptjsonlib.add_node(self.ptjsonlib.create_node_object("phone_numbers", None, None, properties={"phone_numbers": self.result.data["phone_numbers"]}))

        # Find IP Addresses
        if self.extract_types["ip_addresses"]:
            self.result.data["ip_addresses"] = ip_addresses.find_ip_addresses(PAGE_CONTENT)
            self.ptjsonlib.add_node(self.ptjsonlib.create_node_object("ip_addresses", None, None, properties={"ip_addresses": self.result.data["ip_addresses"]}))

        # Find Forms
        if self.extract_types["forms"]:
            soup = self._get_soup(response)
            if soup:
                self.result.data["forms"] = forms.get_forms(soup)
                self.ptjsonlib.add_node(self.ptjsonlib.create_node_object("form", None, None, properties={"forms": self.result.data["forms"]}))

        # Find Absolute URLs
        if any([self.extract_types["external_urls"], self.extract_types["subdomains"], self.extract_types["internal_urls"], self.extract_types["internal_urls_with_parameters"], self.extract_types["insecure_sources"]]):
            self.result.data["abs_urls"] = urls.find_abs_urls(PAGE_CONTENT)
            #self.ptjsonlib.add_node(self.ptjsonlib.create_node_object("abs_urls", None, None, properties={"abs_urls": result_data["abs_urls"]}))

            # Find External URLs | Filters through absolute_urls
            if self.extract_types["external_urls"] or self.extract_types["insecure_sources"]:
                self.result.data['external_urls'] = urls.find_urls_in_response(PAGE_CONTENT, r'(href=|src=)[\'"](.+?)[\'"]', response.url, "external", without_parameters=self.args.without_parameters, abs_urls=self.result.data["abs_urls"])
                if self.extract_types["external_urls"]:
                    self.ptjsonlib.add_node(self.ptjsonlib.create_node_object("external_urls", None, None, properties={"external_urls": self.result.data["external_urls"]}))

            # Find Subdomains | Filters through absolute_urls
            if self.extract_types["subdomains"]:
                self.result.data['subdomains'] = urls.find_urls_in_response(PAGE_CONTENT, r'(href=|src=)[\'"](.+?)[\'"]', response.url, "subdomain", without_parameters=self.args.without_parameters, abs_urls=self.result.data["abs_urls"])
                self.ptjsonlib.add_node(self.ptjsonlib.create_node_object("subdomains", None, None, properties={"subdomains": self.result.data["subdomains"]}))

            # Find Internal URLs | Filters through absolute_urls
            if self.extract_types["internal_urls"] or self.extract_types["internal_urls_with_parameters"] or self.extract_types["insecure_sources"]:
                self.result.data["internal_urls"] = urls.find_urls_in_response(PAGE_CONTENT, r'(href=|src=)[\'"](.+?)[\'"]', response.url, "internal", without_parameters=self.args.without_parameters, abs_urls = self.result.data["abs_urls"])

                # Add to self.ptjsonlib
                if self.extract_types["internal_urls"]:
                    self.ptjsonlib.add_node(self.ptjsonlib.create_node_object("internal_urls", None, None, properties={"internal_urls": self.result.data["internal_urls"]}))

                # Find Internal URLs containing parameters | Filters through internal_urls
                if self.extract_types["internal_urls_with_parameters"]:
                    self.result.data["internal_urls_with_parameters"] = sorted(urls._find_internal_parameters(self.result.data["internal_urls"], group_parameters=self.args.group_parameters), key=lambda k: k['url'])
                    self.ptjsonlib.add_node(self.ptjsonlib.create_node_object("internal_urls_with_parameters", None, None, properties={"internal_urls_with_parameters": self.result.data["internal_urls_with_parameters"]}))
                    if not self.extract_types["internal_urls"]:
                        self.result.data["internal_urls"] = None

            if self.extract_types["insecure_sources"]:
                self.result.data["insecure_sources"] = []
                self.result.data["insecure_sources"].extend(self._find_insecure_sources(all_urls=self.result.data["abs_urls"] + self.result.data["external_urls"] + self.result.data["internal_urls"]))
                self.ptjsonlib.add_node(self.ptjsonlib.create_node_object("insecure_sources", None, None, properties={"insecure_sources": self.result.data["insecure_sources"]}))
                if not self.extract_types["internal_urls"]:
                    self.result.data["internal_urls"] = None
                if not self.extract_types["external_urls"]:
                    self.result.data["external_urls"] = None

        # Extract Google Keys
        if self.extract_types["google_keys"]:
            self.result.data["google_keys"] = google.parse_google_identifiers(response)
            self.ptjsonlib.add_node(self.ptjsonlib.create_node_object("google_keys", None, None, properties={"google_keys": self.result.data["google_keys"]}))

        if self.extract_types["regex"]:
            self.result.data["regex"] = re_search.parse_search_regex(response, self.args.search)

        return self.result

    def _find_insecure_sources(self, all_urls) -> list:
        """Finds URLs loaded via not secure HTTP protocol."""
        insecure_sources = sorted(list(set([u for u in all_urls if re.match(r"http://", u)])))
        return insecure_sources

    def parse_robots_txt(self, response):
        ptprinthelper.ptprint(f"Robots.txt:\n", "TITLE", not self.use_json)
        ptprinthelper.ptprint(response.text.rstrip(), condition=not self.use_json)
        ptprinthelper.ptprint(" ", condition=not self.use_json)
        allow = list({pattern.lstrip() for pattern in re.findall(r"^Allow: ([\S ]*)", response.text, re.MULTILINE)})
        disallow = list({pattern.lstrip() for pattern in re.findall(r"^Disallow: ([\S ]*)", response.text, re.MULTILINE)})
        sitemaps = re.findall(r"[Ss]itemap: ([\S ]*)", response.text, re.MULTILINE)
        test_data = {"allow": allow, "disallow": disallow, "sitemaps": sitemaps}

        parsed_url = urllib.parse.urlparse(response.url)
        internal_urls = []
        for section_header in test_data.values():
            for finding in section_header:
                parsed_finding = urllib.parse.urlparse(finding)
                if not parsed_finding.netloc:
                    full_path = urllib.parse.urlunparse((parsed_url[0], parsed_url[1], parsed_finding[2], "", "", ""))
                else:
                    full_path = finding
                internal_urls.append(full_path)
        return {"url": response.url, "internal_urls": internal_urls}


"""This module contains url extracting functions and related utilities"""

import urllib
import re
import validators

from ptlibs import tldparser

def find_urls_in_file(file_content: str) -> list:
    """Return list of possible links in provided <file_content>."""
    re_abs_urls   = r'(http:\/\/|ftp:\/\/|https:\/\/)([\w_-]+(?:(?:\.[\w_-]+)+))([\w.,@?^=%&:\/~+#-]*[\w@?^=%&\/~+#-])' #(https?:\/\/|ftps?:\/\/)(www\.)?([\w\.]*\.[a-zA-Z]{2,3})(([\w\.\/?=&])*?(?=http))
    re_rel_links  = r"^(?<!<)(\.{1,2}\/?|\/\/|\/)(([\w.]{2,})(\/)?)+$"
    re_html_links = r'(href=|src=)[\'\"](.*)[\'\"]'

    result  = re.findall(re_abs_urls, file_content)
    result += re.findall(re_html_links, file_content)
    result += re.findall(re_rel_links, file_content)

    return sorted(set(''.join(i) if type(i) == tuple else i for i in result))


def get_subdomains_from_list(urls: list) -> list:
    """Return unique list of subdomains from provided <urls>"""
    subdomains = set()
    for url in urls:
        if validators.url(url):
            domain = url2domain(url, with_protocol=False)
        elif validators.domain(url):
            domain = url
        else:
            continue
        subdomains.add(domain)
    return sorted(subdomains)


def find_abs_urls(string: str) -> list:
    """Returns list of absolute urls in <string>"""
    re_abs_url = r'(https?|ftps?)(://[\w\.]*\.[a-zA-Z]{2,3}[?&/#]*[^"\'><\s]*)'
    return sorted(set(''.join(url) for url in re.findall(re_abs_url, string)))


def find_urls_in_response(page_content: str, re_pattern: str, url: str, type_urls: str, without_parameters=None, abs_urls=[]) -> list:
    """Find urls in <page_content> based by applied <re_pattern>"""
    result = []
    domain = url2domain(url)
    all_urls = list({result[1] for result in re.findall(re_pattern, page_content)}) + abs_urls
    for found_url in all_urls:
        if re.match(r'\w*://', found_url) and not re.match(r'https?', found_url):
            continue
        if found_url.startswith("//"):
            all_urls.append(url.split("://")[0] + ":" + found_url)
            continue

        abs_url = rel2abs(found_url, domain)
        parsed_url = urllib.parse.urlparse(abs_url)
        if without_parameters:
            abs_url = urllib.parse.urlunsplit((parsed_url[0], parsed_url[1], parsed_url[2], "", ""))
        if type_urls == "external" and (url2domain(abs_url, False, False) != url2domain(url, False, False)):
            result.append(abs_url)
        if (type_urls == "subdomain") and (url2domain(abs_url, False, False) == url2domain(url, False, False)):
            result.append(parsed_url.netloc)
        if type_urls == "internal" and (url2domain(abs_url, True, False) == url2domain(url, True, False)):
            result.append(abs_url)
    return sorted(list(set(result)), key=str.lower)


def url2domain(url, with_subdomains=True, with_protocol=True) -> str:
    """Returns domain from provided url"""
    parsed_url = tldparser.parse(url)
    protocol = parsed_url.scheme + "://" if with_protocol else ""

    if parsed_url.subdomain:
        parsed_url.subdomain += "."

    if with_subdomains:
        if not parsed_url.suffix: # e.g. ip address
            return protocol + parsed_url.subdomain + parsed_url.domain
        else:
            return protocol + parsed_url.subdomain + parsed_url.domain + "." + parsed_url.suffix
    else:
        return protocol + parsed_url.domain + "." + parsed_url.suffix


def rel2abs(url, domain):
    if not domain.endswith("/") and not url.startswith("/"):
        domain += "/"
    if url.startswith("http://") | url.startswith("https://") | url.startswith("ftp://") | url.startswith("ftps://") | url.startswith("irc://"):
        return url
    else:
        return domain + url


def _find_internal_parameters(internal_urls, group_parameters):
    parsed_urls = []
    for url in internal_urls:
        o = urllib.parse.urlsplit(url)
        if o.query:
            query_list = o.query.split("&")
            parsed_url = urllib.parse.urlunsplit((o[0], o[1], o[2], "", ""))
            result_data = {"url": parsed_url, "parameters": query_list}
            if not group_parameters:
                parsed_urls.append(result_data)
            else:
                if not parsed_urls:
                    parsed_urls.append(result_data)
                    continue
                found = False
                for d in parsed_urls:
                    if parsed_url == d["url"]:
                        d["parameters"].extend(query_list)
                        found = True
                if not found:
                    parsed_urls.append(result_data)
    return parsed_urls


import copy


def get_forms(soup) -> list:
    """Returns parsed page forms"""
    allowed_elements = ["form", "input", "select", "textarea", "label", "button", "datalist", "output"]
    forms = soup.find_all("form")
    forms_result = []
    for form in forms:
        form_elements = strip_form_elements(form.find_all(allowed_elements))
        forms_result.append({"form_name": form.get("name"), "action": form.get("action"), "method": form.get("method"), "form_id": form.get("id"),
        "inputs": [{"tag": ele.name, **ele.attrs} for ele in form_elements["inputs"]], "selects": [{"tag": ele.name, **ele.attrs} for ele in form_elements["selects"]]})
    return forms_result


def pop_value_key_from_form(form):
    """returns form without keys named 'value' (csrf tokens)"""
    form = copy.deepcopy(form)
    if form.get("inputs"):
        for parsed_input in form["inputs"]:
            if "value" in parsed_input.keys():
                parsed_input.pop("value")
    return form


def strip_form_elements(form_elements):
    """strip child elements of parent element"""
    allowed_attrs = ("name", "type", "id", "value")
    result = {"inputs": [], "selects": []}
    for element in form_elements:
        element.attrs = {key: value for key, value in element.attrs.items() if key in allowed_attrs}
        if element.name == "select":
            element.attrs.update({"options": []})
        children = element.findChildren(True, recursive=True)
        for child in children:
            if child.name == "option":
                element.attrs["options"].append(child.get("value", "notfound"))
            else:
                child.unwrap()
        if element.name == "select":
            result["selects"].append(element)
        else:
            result["inputs"].append(element)
    return result


import re


def find_comments(string: str) -> dict:
    """Returns comments from provided string"""
    regex_dict = {"html": [r"<!--[\s\w\W]*?-->"], "css": [r"\/\*[^*]*\*+([^/*][^*]*\*+)*\/"], "js": [r'\/\*[\s\S]+?\*\/']}
    comments = dict()
    for key in regex_dict.keys():
        comments.update({key: []})
        for regex in regex_dict[key]:
            search_result = re.findall(regex, string)
            for item in search_result:
                item = item.strip()
                if item and item not in comments[key]:
                    comments[key].append(item)
    return comments


"""
This module provides a utility to extract Google-related identifiers from HTTP responses.

Supported identifiers include:
- Google Tag Manager ID (GTM-XXXXXX)
- Google Analytics Universal ID (UA-XXXXX-Y)
- Google Analytics 4 ID (G-XXXXXXXX)
- Google Ads Conversion ID (AW-XXXXXXXXX)
- Google Campaign Manager ID (DC-XXXXXX)
- Google AdSense Publisher ID (ca-pub-XXXXXXXXXXXXXXXX)
- Google API Keys (AIza...)

Intended for use in auditing or recon tools to collect telemetry or tracking IDs embedded in webpages.
"""

import re
from requests import Response
from ptlibs import ptprinthelper


def parse_google_identifiers(response: Response) -> None:
    """
    Parses an HTTP response for various Google-related tracking or service identifiers.

    Args:
        response (Response): The HTTP response object from the `requests` library.

    Scans the response body for the following patterns:
        - GTM (Google Tag Manager)
        - UA (Universal Analytics)
        - G- (Google Analytics 4)
        - AW- (Google Ads conversion ID)
        - DC- (Campaign Manager ID)
        - ca-pub / ca-ads (AdSense publisher ID)
        - API keys starting with "AIza"

    If found, prints the identifiers grouped by type.
    Relies on the `ptprinthelper.ptprint` function and `self.args.json` flag for conditional output.
    """

    #ptprinthelper.ptprint(f"Google identifiers", "TITLE", condition=not self.args.json, colortext=True, newline_above=True)

    regulars = {
        "Google Tag Manager ID": r"(GTM-[A-Z0-9]{6,9})",
        "Google Analytics Universal ID": r"(UA-\d{4,10}-\d+)",
        "Google Analytics 4": r"(G-[A-Z0-9]{8,12})",
        "Google Ads Conversion ID": r"(AW-\d{9,12})",
        "Google Campaign Manager ID": r"(DC-\d{6,10})",
        "Google AdSense Publisher ID" : r"(ca-pub-\d{16})|(ca-ads-\d{16})",
        "Google API Keys": r"AIza[0-9A-z_\-\\]{35}",
    }

    found_identifiers = {}
    for key, regex in regulars.items():
        matches = re.findall(regex, response.text)
        matches = [m[0] if isinstance(m, tuple) else m for m in matches]
        matches = sorted(set(matches))
        if matches:
            found_identifiers[key] = matches

    return found_identifiers


from collections.abc import Callable

def parse_extract_flags(args, extract_str: str, on_error: Callable[[str], None]) -> dict:
    """
    Parses a short extract string (e.g., "ESCF") and returns a dictionary
    indicating which extraction types are enabled.

    Args:
        extract_str (str): String containing one or more valid extract type characters.
        on_error (Callable[[str], None]): Function to call on error with the
            error message as argument. Must be provided.

    Returns:
        dict: Dictionary with extraction type keys and `True` or `None` as values.
    """

    allowed_letters = {
        "E": "emails",
        "S": "subdomains",
        "C": "comments",
        "F": "forms",
        "U": "internal_urls",
        "X": "external_urls",
        "P": "phone_numbers",
        "M": "metadata",
        "N": "insecure_sources",
        "I": "ip_addresses",
        "Q": "internal_urls_with_parameters",
        "R": "regex",
        "K": "google_keys",
        "A": "all"
    }

    extract_types = {value: None for value in allowed_letters.values()}
    extract_str = extract_str.upper()

    # Check for invalid characters
    invalid = [c for c in extract_str if c not in allowed_letters]

    if invalid:
        msg = (
            f"Invalid parameter(s) '{''.join(invalid)}' in --extract argument.\n"
            f"Allowed characters:\n" +
            "\n".join([f"  {k} - {v}" for k, v in allowed_letters.items()])
        )
        on_error(msg)
        return extract_types  # or raise, pokud chceš proces ukončit v on_error

    if "A" in extract_str:
        # Enable all types; regex depends on args.search
        for key in extract_types:
            extract_types[key] = True
        if not args.search:
            extract_types["regex"] = None
    else:
        # Enable only explicitly specified types
        for char in extract_str:
            key = allowed_letters[char]
            extract_types[key] = True


    # Check regex requirement
    if args.search is not None and "R" not in extract_str and "A" not in extract_str:
        on_error("You must include 'R' in --extract when using --search.")

    # Enable regex if present and search is provided
    if "R" in extract_str:
        if args.search is None:
            on_error("You specified 'R' in --extract but did not provide --search.")
        else:
            extract_types["regex"] = True

    return extract_types



from modules import forms
from ptlibs import ptprinthelper
from ptlibs.ptprinthelper import get_colored_text
from modules.metadata import write_metadata_to_csv

def print_result(list_of_results: list, args):
    """
    Prints or writes processed results in a structured format based on provided arguments.

    The function handles:
        - Conditional printing of logs (including errors)
        - Grouped or complete summary views
        - Filtering and formatting of different result types (internal URLs, metadata, forms, comments, Google keys, etc.)
        - Writing output to files, including separate files per data type if specified

    Parameters:
        list_of_results (list): A list of result objects, each containing 'data' and 'logs' attributes.
        args: An object (typically argparse.Namespace) containing configuration flags such as:
            - json (bool): Whether to suppress console output for JSON output.
            - grouping (bool): Whether to print results grouped by URL.
            - grouping_complete (bool): Whether to print a merged summary of all results.
            - output (str or None): Path to the output file.
            - output_parts (bool): Whether to write separate files for each data type.
            - file_handle: Internal handle for writing to files.

    Returns:
        None: Results are printed to the console or written to files based on arguments.
    """

    handlers = {
        "internal_urls_with_parameters": _print_parsed_urls,
        "metadata": _print_metadata,
        "forms": _print_forms,
        "comments": _print_comments,
        "google_keys": _print_google_keys,
    }
    if not len(list_of_results) > 1:
        for r in list_of_results:
            ptprinthelper.ptprint(" ", "TEXT", condition=not args.json)
            for log, msg_type in r.logs:
                ptprinthelper.ptprint(log, msg_type, condition=not args.json)
                if msg_type.upper() == "ERROR":
                    pass
    else:
        if args.grouping_complete:
            ptprinthelper.ptprint("\n", "TEXT", condition=not args.json)
            ptprinthelper.ptprint(get_colored_text(f"SUMMARY (UNIQUE VALUES)\n", color="TITLE") + f"{'='*len('Summary (unique values)')}", "TEXT", condition=not args.json, newline_above=True)

    list_of_results = [l.data for l in list_of_results]
    titles = [i for i in {k: v for k, v in list_of_results[0].items() if v is not None}.keys()]
    tested_things = [key for key, value in args.extract.items() if value is True]
    titles = [t for t in titles if t in tested_things]
    args.file_handle = None
    if args.output and not args.output_parts:
        args.file_handle = open(args.output, "w")

    for index, title in enumerate(titles):
        if title in ["url", "abs_urls"]:
            continue

        if args.output_parts:
            args.file_handle = _get_handle(title, args)

        if args.grouping_complete:
            merged_result = _get_data_type(title)

        _print_title(title, args, index)
        for index, result_dictionary in enumerate(list_of_results):
            if args.grouping:
                _print_current_url(list_of_results, index, args)

            if args.grouping_complete:
                merged_result = _fill_merged_result(merged_result, list_of_results, index, title)

            # if not -gc
            else:
                if title == "internal_urls_with_parameters":
                    _print_parsed_urls(list_of_results, index, args)
                elif title == "metadata":
                    _print_metadata(list_of_results, index, args)
                elif title == "forms":
                    _print_forms(list_of_results, index, args)
                elif title == "comments":
                    _print_comments(list_of_results, index, args)
                elif title == "google_keys":
                    _print_google_keys(list_of_results, index, args)
                elif result_dictionary.get(title):
                    ptprinthelper.ptprint("\n".join(result_dictionary[title]), filehandle=args.file_handle, condition=not args.json)
                else:
                    ptprinthelper.ptprint("Not found", "", filehandle=args.file_handle, condition=not args.json)
                _get_endl(list_of_results, index, args)

        if args.grouping_complete:
            # if -gc provided
            if title == "internal_urls_with_parameters":
                merged_result["internal_urls_with_parameters"] = sorted(merged_result["internal_urls_with_parameters"], key=lambda k: k['url'])
                _print_parsed_urls([merged_result], index=0, args=args)
            elif title == "metadata":
                _print_metadata([merged_result], index=0, args=args)
                if args.output and len(list_of_results) > 1:
                    # final -gc result where we have all metadata collected.
                    final = [res.get("metadata", {}) for res in list_of_results]
                    write_metadata_to_csv(final, args.output)

            elif title == "forms":
                _print_forms([merged_result], index=0, args=args)
            elif title == "comments":
                _print_comments([merged_result], index=0, args=args)
            elif title == "google_keys":
                _print_google_keys([merged_result], index=0, args=args)
            elif title == "regex":
                if not merged_result["regex"].keys():
                    ptprinthelper.ptprint("Not found", "", filehandle=args.file_handle, condition=not args.json)
                for text, urls in merged_result["regex"].items():
                    if len(list_of_results) > 1 and args.very_verbose:
                        max_len = max(len(text) for text in merged_result["regex"].keys())
                        padding = " " if (max_len - len(text)) > 50 else " " * (max_len - len(text) +1)
                        ptprinthelper.ptprint(f"{text.strip()}{padding}", "", condition=not args.json, end="\n" if not args.very_verbose else "")
                        if args.very_verbose:
                            ptprinthelper.ptprint(ptprinthelper.get_colored_text(f"({', '.join(sorted(urls))})", color="ADDITIONS"), "", condition=not args.json)
                    else:
                        ptprinthelper.ptprint(f"{text}", "", condition=not args.json)

                # print dynamic title with unique urls
                if len(list_of_results) > 1:
                    ptprinthelper.ptprint(f"{ptprinthelper.get_colored_text('REGEX URLS', color="TITLE")}\n{'-'*len('REGEX URLS')}", "TEXT", condition=not args.json, newline_above=True)
                    _all_urls = set().union(*merged_result["regex"].values())
                    ptprinthelper.ptprint("\n".join(sorted(_all_urls)), "", condition=not args.json)
            elif merged_result:
                ptprinthelper.ptprint("\n".join(sorted(set(merged_result))), "", end=_check_if_next(list_of_results, index)+"\n", filehandle=args.file_handle, condition=not args.json)
            else:
                ptprinthelper.ptprint("Not found", "", filehandle=args.file_handle, condition=not args.json)

        if args.output_parts:
            args.file_handle.close()

    if args.file_handle:
        args.file_handle.close()


def _get_handle(title, args):
    parsed_file_name = args.output.rsplit(".", 1)
    file_extension = "." + parsed_file_name[-1] if len(parsed_file_name) > 1 else ""
    handle = open(f"{parsed_file_name[0]}_{title.upper()}{file_extension}", "w")
    return handle



def _print_google_keys(list_of_results, index, args):
    result_dict = list_of_results[index].get("google_keys")
    if result_dict:
        for key, value in result_dict.items():
            ptprinthelper.ptprint(f"{key}:", "TEXT", condition=not args.json, indent=0)
            ptprinthelper.ptprint("\n    ".join(value), "TEXT", condition=not args.json, indent=4)
            ptprinthelper.ptprint(f" ", "TEXT", condition=not args.json)#
    else:
        ptprinthelper.ptprint("Not found", "TEXT", not args.json, filehandle=args.file_handle)

def _print_title(title, args, index=0):
    if title == "regex":
        title = "REGEX SEARCH"
    if args.file_handle and index > 0:
        args.file_handle.write("\n")
    #ptprinthelper.ptprint(f'\n{ptprinthelper.get_colored_text(title.upper().replace("_", " "), color="TITLE")}{"" * len(title)}', "", filehandle=args.file_handle, condition=not args.json)
    ptprinthelper.ptprint(f'\n{ptprinthelper.get_colored_text(title.upper().replace("_", " "), color="TITLE")}\n{"-" * len(title)}', "", filehandle=args.file_handle, condition=not args.json)
    #ptprinthelper.ptprint(f'{ptprinthelper.get_colored_text(title.upper().replace("_", " "), color="TITLE")}', "TITLE", filehandle=args.file_handle, condition=not args.json, newline_above=True)


def _get_endl(list_of_results, index, args):
    """Add space if not last record"""
    if args.grouping and list_of_results[index] != list_of_results[-1]:
        ptprinthelper.ptprint(f" ", "", filehandle=args.file_handle, condition=not args.json)


def _check_if_next(list_of_results, index):
    try:
        endl = "\n" if list_of_results[index+1] else ""
    except Exception:
        endl = ""
    return endl


def _print_parsed_urls(list_of_results, index, args):
    if not list_of_results[index]["internal_urls_with_parameters"]:
        ptprinthelper.ptprint("Not found", "", not args.json, filehandle=args.file_handle)
    elif type(list_of_results[index]["internal_urls_with_parameters"]) == str: # If error msg (eg. cannot search file)
        ptprinthelper.ptprint(list_of_results[index]["internal_urls_with_parameters"], "", filehandle=args.file_handle, condition=not args.json)
    else:
        for url in list_of_results[index]["internal_urls_with_parameters"] or []:
            ptprinthelper.ptprint(f"URL: {url['url']}", "", filehandle=args.file_handle, condition=not args.json,)
            ptprinthelper.ptprint(f"Parameters:", "", filehandle=args.file_handle, condition=not args.json)
            for parameter in url['parameters']:
                ptprinthelper.ptprint(f'          {parameter}', "", condition=not args.json, filehandle=args.file_handle, end="\n")
            if url != list_of_results[index]["internal_urls_with_parameters"][-1]: # Add space if not last
                ptprinthelper.ptprint(f" ", "", filehandle=args.file_handle, condition=not args.json)


def _print_comments(list_of_results, index, args):
    if not any(list_of_results[index]["comments"].values()):
        ptprinthelper.ptprint(f"Not found", "", filehandle=args.file_handle, condition=not args.json)
    else:
        for key in list_of_results[index]["comments"].keys():
            if list_of_results[index]["comments"][key]:
                #ptprinthelper.ptprint(ptprinthelper.get_colored_text(key.upper(), "TITLE"), filehandle=args.file_handle, condition=not args.json)
                ptprinthelper.ptprint('\n'.join(list_of_results[index]["comments"][key]), "", filehandle=args.file_handle, condition=not args.json)


def _print_metadata(list_of_results, index, args):
    longest_key, _ = max(list_of_results[index]["metadata"].items(), key=lambda x: len(x[0]), default=("x", None))
    for key, value in list_of_results[index]["metadata"].items():
        if key == "SourceFile": continue
        if type(value) is list:
            value = ', '.join(map(str, value))
        else:
            value = str(value).replace("\n", "\\n")
        ptprinthelper.ptprint(f"{key}{' '*(len(longest_key)-len(key))}: {', '.join(value) if type(value) is list else value}", "", filehandle=args.file_handle, condition=not args.json)


def _print_forms(list_of_results, index, args):
    if args.grouping_complete:
        if not list_of_results[0]["forms"]:
            ptprinthelper.ptprint("Not found", "", filehandle=args.file_handle, condition=not args.json)
        else:
            _pretty_print_forms(list_of_results, index, args)
    else:
        if not list_of_results[index]["forms"]:
            ptprinthelper.ptprint("Not found", "", filehandle=args.file_handle, condition=not args.json)
        elif type(list_of_results[index]["forms"]) == str:
            ptprinthelper.ptprint(list_of_results[index]["forms"], "", filehandle=args.file_handle, condition=not args.json)
        else:
            _pretty_print_forms(list_of_results, index, args)


def _print_current_url(list_of_results, index, args):
    ptprinthelper.ptprint(ptprinthelper.get_colored_text(f"{list_of_results[index]['url']}", 'INFO'), "", not args.json, filehandle=args.file_handle)


def _get_data_type(title: str):
    """
    Returns an empty data structure appropriate for the given title/type.

    Parameters:
        title (str): The type of data (e.g., 'metadata', 'forms', 'comments', 'google_keys', etc.)

    Returns:
        dict or set: An empty data structure for aggregating results of this type.
            - 'metadata' -> {'metadata': {}}
            - 'internal_urls_with_parameters' -> {'internal_urls_with_parameters': []}
            - 'forms' -> {'forms': []}
            - 'comments' -> {'comments': {'html': [], 'js': [], 'css': []}}
            - 'google_keys' -> {'google_keys': {}}
            - Any other title -> set()
    """
    if title == "metadata":
        merged_result = {"metadata": {}}
    elif title == "internal_urls_with_parameters":
        merged_result = {"internal_urls_with_parameters": []}
    elif title == "forms":
        merged_result = {"forms": list()}
    elif title == "comments":
        merged_result = {"comments": {"html": [], "js": [], "css": []}}
    elif title == "google_keys":
        merged_result = {"google_keys": {}}
    elif title == "regex":
        merged_result = {"regex": {}}
    else:
        merged_result = set()
    return merged_result


def _pretty_print_forms(list_of_results, index, args):
    for idx, form in enumerate(list_of_results[index]["forms"]):
        for key, value in form.items():
            space = 0 if key == "form_name" else 9
            if value == '':
                value = "''"
            if key in ["inputs", "selects"]:
                if form[key]:
                    ptprinthelper.ptprint(" ", "", filehandle=args.file_handle, condition=not args.json)
                if not form[key]:
                    continue
                ptprinthelper.ptprint(f"{' '*space}{key.title()}:", "", filehandle=args.file_handle, condition=not args.json)
                space += len(key)
                for idx2, dictionary in enumerate(form[key]):
                    for key2, value2 in dictionary.items():
                        if not value2 and value2 is not None:
                            value2 = "''"
                        if key2 == "options":
                            ptprinthelper.ptprint(f"{' '*space}{key2.title()}:", "", filehandle=args.file_handle, condition=not args.json)

                            space += len(key)
                            for option in dictionary[key2]:
                                if not option and option is not None:
                                    option = "''"
                                ptprinthelper.ptprint(f"{' '*space}{option}", "", filehandle=args.file_handle, condition=not args.json)
                            if args.grouping_complete and form != list_of_results[index]["forms"][-1]:
                                ptprinthelper.ptprint(" ", "", filehandle=args.file_handle, condition=not args.json)
                        else:
                            ptprinthelper.ptprint(f"{' '*space}{key2.title()}: {value2}", "", filehandle=args.file_handle, condition=not args.json)
                    if idx2+1 != len(form[key]):
                        ptprinthelper.ptprint(" ", "", filehandle=args.file_handle, condition=not args.json)
            else:
                ptprinthelper.ptprint(f"{' '*space}{key.title().replace('_',' ')}: {value}", "", filehandle=args.file_handle, condition=not args.json)
        if idx+1 != len(list_of_results[index]["forms"]):
            ptprinthelper.ptprint(" ", "", filehandle=args.file_handle, condition=not args.json)


def _fill_merged_result(merged_result, list_of_results, index, title):
    """
    Merges the data from a single result into the cumulative merged_result based on the title/type.

    Parameters:
        merged_result (dict or set): The current aggregated data structure for the title.
        list_of_results (list): List of all results (each is a dict or BaseResult.data) to merge from.
        index (int): Index of the current result in list_of_results to merge.
        title (str): The type/key of data to merge (e.g., 'metadata', 'forms', 'comments', etc.).

    Returns:
        dict or set: Updated merged_result containing merged values from the current result.

    Notes:
        - For 'metadata', it merges keys and lists without duplicates.
        - For 'forms', 'internal_urls_with_parameters', 'comments', and 'google_keys',
          it extends or appends the corresponding lists/dictionaries.
        - For all other titles, it assumes a set and adds new items.
        - This function does not modify the original list_of_results.
    """
    if title == "internal_urls_with_parameters":
        for result_dict in list_of_results[index]["internal_urls_with_parameters"] or []:
            if result_dict not in merged_result["internal_urls_with_parameters"]:
                merged_result["internal_urls_with_parameters"].append(result_dict)

    elif title == "metadata":
        if list_of_results[index].get("metadata"):
            for key, value in list_of_results[index]["metadata"].items():
                if key not in merged_result["metadata"]:
                    merged_result["metadata"][key] = value

                elif isinstance(merged_result["metadata"][key], list):
                    if isinstance(value, list):
                        merged_result["metadata"][key].extend([val for val in value if val not in merged_result["metadata"][key]])
                    else:
                        if value not in merged_result["metadata"][key]: merged_result["metadata"][key].append(value)
                else:
                    if isinstance(value, list):
                        merged_result["metadata"][key] = [merged_result["metadata"][key]]
                        merged_result["metadata"][key].extend([val for val in value if val not in merged_result["metadata"][key]])
                    else:
                        merged_result["metadata"][key] = [merged_result["metadata"][key]]
                        if value not in merged_result["metadata"][key]: merged_result["metadata"][key].append(value)

    elif title == "forms":
        for form in list_of_results[index]["forms"] or []:
            form_without_value_keys = forms.pop_value_key_from_form(form)
            if form_without_value_keys not in merged_result["forms"]:
                merged_result["forms"].append(form_without_value_keys)

    elif title == "comments":
        for key, value_list in (list_of_results[index].get("comments") or {}).items():
            merged_result["comments"][key].extend(value_list)
            merged_result["comments"][key] = list(set(merged_result["comments"][key]))

    elif title == "google_keys":
        merged_result.setdefault("google_keys", {})
        for key, value_list in (list_of_results[index].get("google_keys") or {}).items():
            merged_result["google_keys"].setdefault(key, []).extend(value_list)

    elif title == "regex":
        found_texts = list_of_results[index].get("regex", [])
        url = list_of_results[index].get("url", "Unknown URL")
        for text in found_texts:
            if text not in merged_result["regex"]:
                merged_result["regex"][text] = set()
            merged_result["regex"][text].add(url)

    else:
        if list_of_results[index].get(title):
            for i in list_of_results[index][title]:
                merged_result.add(i)

    return merged_result



import re

from bs4 import BeautifulSoup

from ptlibs import ptprinthelper
from modules import metadata, emails, comments, forms, phone_numbers, ip_addresses, urls

from modules.results.base_result import BaseResult

class FileScraper:
    def __init__(self, args, things_to_extract: dict, ptjsonlib: object, lock: object):
        self.args = args
        self.extract_types: dict = things_to_extract
        self.lock = lock
        self.ptjsonlib = ptjsonlib

    def process_file(self, path_to_local_file: str) -> dict:
        """Returns extracted info from <filepath>"""
        self.result = BaseResult(source=path_to_local_file)
        try:
            import magic
            import mimetypes
            mimetype       = mimetypes.guess_type(path_to_local_file)[0]
            content_type   = magic.from_file(path_to_local_file)
        except Exception as e:
            self.ptjsonlib.end_error("Dependency not found, please install: sudo apt install libmagic1", self.args.json)

        is_readable    = True if re.findall("text", content_type.lower()) else False
        file_extension = path_to_local_file.split("/")[-1].split(".")[-1] if "." in path_to_local_file.split("/")[-1] else None

        self.result.logs.append([f"Provided source: {path_to_local_file.split('/')[-1]}", "INFO"])
        self.result.logs.append([f"Extension: {file_extension}", "INFO"])
        self.result.logs.append([f"Source-Type: {content_type}", "INFO"])
        self.result.logs.append([f"MIME Type: {mimetype}", "INFO"])

        result = self._scrape_file(path_to_local_file, is_readable, mimetype)
        return result

    def _scrape_file(self, path_to_local_file: str, is_readable, mimetype) -> dict:
        """Scrapes extract_types from <filepath>"""
        self.result.data: dict = {"url": path_to_local_file.rsplit("/")[-1], "metadata": {}, "emails": [], "phone_numbers": [], "ip_addresses": [], "abs_urls": [], "internal_urls": [], "internal_urls_with_parameters": [], "external_urls": [], "insecure_sources": [], "subdomains": [], "forms": [], "comments": {}}
        if self.extract_types["metadata"]:
            extracted_metadata = metadata.MetadataExtractor(args=self.args, ptjsonlib=self.ptjsonlib).get_metadata(path_to_local_file=path_to_local_file)
            self.result.data["metadata"] = extracted_metadata
            if len(extracted_metadata.keys()) == 1 and "ExifTool" in list(extracted_metadata)[0]:
                self.result.data["metadata"] = {list(extracted_metadata)[0].split("ExifTool:")[-1]: list(extracted_metadata.values())[0]}

        if not is_readable:
            return self.result

        with open(path_to_local_file, "rb") as file:
            file_content = str(file.read())

            if self.extract_types["emails"]:
                self.result.data["emails"] = emails.find_emails(file_content)

            if self.extract_types["comments"]:
                self.result.data["comments"] = {}

            if self.extract_types["phone_numbers"]:
                self.result.data["phone_numbers"] = phone_numbers.find_phone_numbers(file_content)

            if self.extract_types["ip_addresses"]:
                self.result.data["ip_addresses"] = ip_addresses.find_ip_addresses(file_content)

            if any([self.extract_types["internal_urls"], self.extract_types["external_urls"], self.extract_types["internal_urls_with_parameters"], self.extract_types["subdomains"]]):
                self.result.data["external_urls"] = urls.find_urls_in_file(file_content)

            if self.extract_types["subdomains"]:
                self.result.data["subdomains"] = urls.get_subdomains_from_list(self.result.data["external_urls"])

            if self.extract_types["internal_urls_with_parameters"]:
                self.result.data["internal_urls_with_parameters"] = dict()
                """
                if args.grouping_complete:
                    self.result.data["internal_urls_with_parameters"] = dict()
                else:
                    self.result.data["internal_urls_with_parameters"] = "Not a HTML file"
                """

            if self.extract_types["forms"]:
                if str(mimetype) in ["text/html"]:
                    soup = self._get_soup(file_content)
                    self.result.data["forms"] = forms.get_forms(soup)
                else:
                    self.result.data["forms"] = {}
                    """
                    if args.grouping_complete:
                        self.result.data["forms"] = dict()
                    else:
                        self.result.data["forms"] = "Not a HTML file"
                    """
        return self.result

    def _get_soup(self, string):
        if "<!ENTITY".lower() in string.lower():
            ptprinthelper.ptprint(f"Forbidden entities found", "ERROR", not self.args.json, colortext=True)
            return False
        else:
            soup = BeautifulSoup(string, features="lxml")
            bdos = soup.find_all("bdo", {"dir": "rtl"})
            for item in bdos:
                item.string.replace_with(item.text[::-1])
            return soup



import json

class BaseResult:
    """
    A result class for scraping operations, supporting specific result types such as website pages, robots.txt,
    sitemaps, and files.

    This class stores the source of the data (either URL for HTTP or file path for file-based operations),
    a success flag, logs, and the actual scraped data.

    Attributes:
        source (str): The source of the scraped data. For HTTP, it's the URL; for file scraping, it's the file path.
        success (bool): Indicates whether the scrape was successful.
        logs (list): A list of log messages or errors encountered during the scraping operation.
        data (dict): A dictionary containing the scraped data, such as emails, URLs, or other relevant information.
        result_type (str): The type of the result, which can be one of the following:
            - 'website': Classic website scrape (e.g., HTML page response).
            - 'robots_txt': Scraped robots.txt file (contains crawl instructions for web crawlers).
            - 'sitemap': Scraped sitemap file (typically XML containing URLs of a website).
            - 'file': File-based scraping (e.g., local file scraping, such as text or CSV).
    """

    def __init__(self, source: str, success: bool = False, logs: list = None, data: dict = None, result_type: str = 'website'):
        """
        Initializes the BaseResult instance.

        Args:
            source (str): The source of the scraped data. Can be a URL for HTTP scraping or a file path for file scraping.
            success (bool): Whether the scrape was successful or not.
            logs (list): List of logs or error messages encountered during scraping.
            data (dict): A dictionary containing the actual scraped data (e.g., emails, URLs).
            result_type (str): The type of the result, which can be one of the following:
                - 'website': Classic website scrape (e.g., HTML page response).
                - 'robots_txt': Scraped robots.txt file.
                - 'sitemap': Scraped sitemap file.
                - 'file': File-based scraping (e.g., local file scraping).
        """
        self.source = source
        self.success = success
        self.logs = logs if logs else []
        self.data = data if data else {}

    def __repr__(self):
        return str(json.dumps(self.to_dict(), indent=4))

    def to_dict(self):
        """
        Converts the BaseResult object to a dictionary format.

        Returns:
            dict: A dictionary representation of the scrape result.
        """
        return {
            "source": self.source,
            "success": self.success,
            "logs": self.logs,
            "data": self.data
        }


import re
import os


def find_emails(string: str):
    """Returns list of found emails from provided string"""
    email_regex = r'[\w\.-]+@[\w\.-]+\.[a-zA-Z]{2,3}'
    return sorted(set(re.findall(email_regex, string)))



import re


def find_phone_numbers(string: str):
    """Returns list of found phone numbers from <string>"""
    phone_regex = r"(?<![\w\/=\:-])(\(?\+\d{3}\)?[ -]?)?(\d{3}[ -]?)(\d{2,3}[ -]?)(\d{3}|\d{2} \d{2})(?![\w\"\'\/\\.])"
    search_result = re.findall(phone_regex, string)
    return sorted(set([''.join(tuple_result) for tuple_result in search_result]))


import os
import csv
import tempfile
import exiftool
import shutil

class MetadataExtractor:
    def __init__(self, args, ptjsonlib):
        self.args = args
        self.ptjsonlib = ptjsonlib
        self.exif_executable = shutil.which("exiftool")
        if not self.exif_executable:
            self.ptjsonlib.end_error("ExifTool not found. install first via 'sudo apt install exiftool' command.", self.args.json)

    def get_metadata(self, response=None, path_to_local_file=None) -> dict:
        if response is not None:
            tmp = tempfile.NamedTemporaryFile()
            with open(tmp.name, 'wb') as f:
                f.write(response.content)
            with exiftool.ExifTool(executable=self.exif_executable) as exif_tool:
                result_dict = exif_tool.execute_json(tmp.name)[0]

        elif path_to_local_file:
            with exiftool.ExifTool(executable=self.exif_executable) as exif_tool:
                result_dict = exif_tool.execute_json(path_to_local_file)[0]

        blacklisted_keys = ["SourceFile", "ExifTool:ExifToolVersion", "File:FileName", "File:Directory", "File:FileSize", "File:FileModifyDate", "File:FileInodeChangeDate", "File:FilePermissions", "File:FileAccessDate", "File:FileType", "File:FileTypeExtension", "File:MIMEType"]
        result_dict = {k: v for k, v in result_dict.items() if k not in blacklisted_keys and v}
        file_identifier = response.url if response is not None else path_to_local_file
        result_dict.update({"SourceFile": file_identifier})

        """
        if self.args.output:
            self._write_metadata_to_csv(result_dict, file_identifier)
        """
        return result_dict

def write_metadata_to_csv(metadata_list, output_prefix):
    """
    Converts a list of metadata dictionaries to a CSV file.
    Ensures 'SourceFile' is the first column.
    """
    csv_path = f"{output_prefix}-metadata.csv"

    # Collect all unique keys across all dictionaries
    all_keys = set()
    for meta in metadata_list:
        all_keys.update(meta.keys())

    # Reorder keys: SourceFile first
    all_keys = ["SourceFile"] + [k for k in all_keys if k != "SourceFile"]

    # Write to CSV
    with open(csv_path, mode='w', newline='', encoding='utf-8') as csvfile:
        writer = csv.DictWriter(csvfile, fieldnames=all_keys, quotechar='"', quoting=csv.QUOTE_ALL)
        writer.writeheader()
        for meta in metadata_list:
            row = {}
            for k in all_keys:
                value = meta.get(k, "")
                if isinstance(value, list):
                    value = "; ".join(str(v).replace(",", ";") for v in value)
                else:
                    value = str(value).replace(",", ";")
                row[k] = value
            writer.writerow(row)